from typing import Literal


Periods_T = Literal["1h", "24h", "8hg", "24hg", "1m", "1y", "1yaot40", "5yaot40"]

TimeSpans_T = Literal[
    "currentday",
    "currentmonth",
    "currentyear",
    "lastday",
    "lastweek",
    "lastmonth",
    "lastyear",
    "all",
    "custom",
]

StationGroup_T = Literal[
    "all", "background", "industrial", "meteorology", "ozone", "special", "traffic", "pollution"
]

GroupCodes_T = Literal["meteorology", "pollution"]



from datetime import datetime
from typing import Annotated, Any, Self

from pydantic import BaseModel, field_serializer, Field

from ._typing import GroupCodes_T, Periods_T, TimeSpans_T, StationGroup_T


class BaseParams(BaseModel):
    pass

class _ActiveParam(BaseParams):
    active: bool | None = None

class _HiddenParam(BaseParams):
    include_hidden: bool | None = None

class ListParams(_ActiveParam, _HiddenParam):
    pass

class ItemListModel(BaseParams):
    @classmethod
    def from_response(cls, result: list[Any]) -> list[Self]:
        return [cls(**item) for item in result]



class Core(ItemListModel):
    code: str
    active: bool
    group: StationGroup_T
    name: str
    short_name: Annotated[str, Field(validation_alias="shortName")]
    description: str | None
    default_period: Annotated[str, Field(validation_alias="defaultPeriod")]
    periods: list[str]
    decimal_points: Annotated[int, Field(validation_alias="decimalPoints")]
    components: list[str]


class Threshold(ItemListModel):
    active: bool
    code: str
    component: str
    core: str
    name: str
    description: str | None
    value: float
    max_count: Annotated[int | None, Field(validation_alias="maxCount")]
    start: datetime | None


class DatetimeRangeParams(BaseParams):
    start: datetime | None = None
    end: datetime | None = None

    @field_serializer("start", "end")
    def custom_dt(self, dt: datetime) -> str:
        return dt.strftime("%d.%m.%Y %H:%M")


class ComponentParams(_ActiveParam):
    group_code: GroupCodes_T | None


class Station(ItemListModel):
    name: str
    code: str
    code_eu: Annotated[str, Field(validation_alias="codeEu")]
    address: str
    lat: str
    lng: str
    active: bool
    stationgroups: list[str]
    measuring_start: Annotated[
        datetime | None, Field(validation_alias="measuringStart")
    ]
    measuring_end: Annotated[datetime | None, Field(validation_alias="measuringEnd")]
    measuring_height: Annotated[int | None, Field(validation_alias="measuringHeight")]
    url: str
    information: str | None
    components: list[str]
    active_components: Annotated[list[str], Field(validation_alias="activeComponents")]
    partials: list
    lqis: list[str]
    exceeds: list[str]


class StationData(ItemListModel):
    datetime: datetime
    station: str
    core: str
    component: str
    period: Periods_T
    value: int


class ComponentDataParams(DatetimeRangeParams):
    stationgroup: str | None = None
    timespan: TimeSpans_T | None = None


class ComponentData(ItemListModel):
    datetime: datetime
    station: str
    core: str
    component: str
    period: Periods_T
    value: float | None


class Component(ItemListModel):
    code: str
    core: str
    period: str
    description: str | None
    unit: str
    min: int
    max: int
    decimal_points: Annotated[int, Field(validation_alias="decimalPoints")]


class StationDataParams(DatetimeRangeParams):
    period: Periods_T
    timespan: TimeSpans_T
    core: str | None



from datetime import datetime
from collections.abc import Callable
import json
import time
from typing import Any, Protocol
import logging

import hishel
import httpx
from ratelimit import limits, sleep_and_retry

from httpx._types import QueryParamTypes

from ._models import (
    BaseParams,
    Component,
    ComponentParams,
    ComponentData,
    ComponentDataParams,
    Core,
    Station,
    StationDataParams,
    Threshold,
    ListParams,
    StationData,
)

from ._typing import GroupCodes_T, TimeSpans_T, Periods_T

__all__ = ["LuftmessnetzClient"]

logger = logging.getLogger(__name__)
LOG_FMT = "%(levelname)-10s %(message)s"
MAX_RETRIES = 8


class LuftmessnetzClient:
    BASE_URL = "https://hamburg.luftmessnetz.de/api"

    def __init__(self) -> None:
        self._client = hishel.CacheClient()

    def get_components(
        self,
        *,
        active: bool | None = None,
        group_code: GroupCodes_T | None = None,
    ) -> list[Component]:
        """Retrieve component's metadata."""
        params = ComponentParams(active=active, group_code=group_code)
        return Component.from_response(self._get("components", params))

    def get_component_data(
        self,
        code: str,
        *,
        stationgroup: str | None = None,
        timespan: TimeSpans_T | None = None,
        start: datetime | None = None,
        end: datetime | None = None,
    ) -> list[ComponentData]:
        params = ComponentDataParams(
            stationgroup=stationgroup, timespan=timespan, start=start, end=end
        )
        result = self._get(f"components/{code}/data", params)
        return ComponentData.from_response(result)

    def get_cores(self) -> list[Core]:
        """Retrieve core metadata."""
        return Core.from_response(self._get("cores"))

    def get_thresholds(self) -> list[Threshold]:
        """Retrieve the thresholds for each each core."""
        return Threshold.from_response(self._get("exceeds"))

    def get_stations(
        self,
        *,
        active: bool | None = None,
        include_hidden: bool | None = None,
    ) -> list[Station]:
        """Retrieve a list of all measuring stations and their metadata."""
        params = ListParams(active=active, include_hidden=include_hidden)
        return Station.from_response(self._get("stations", params=params))

    def get_station(self, code: str) -> Station:
        """Retrieve metadata of a single station."""
        return Station(**self._get(f"stations/{code}"))

    def get_station_data(
        self,
        station_code: str,
        *,
        period: Periods_T,
        timespan: TimeSpans_T,
        start: datetime | None = None,
        end: datetime | None = None,
        core: str | None = None,
    ) -> list[StationData]:
        params = StationDataParams(
            period=period, timespan=timespan, start=start, end=end, core=core
        )
        return StationData.from_response(
            self._get(f"stations/{station_code}/data", params)
        )

    @sleep_and_retry
    @limits(120, period=60)
    def _get(self, endpoint: str, params: BaseParams | None = None) -> Any:
        params_ = params.model_dump(exclude_none=True) if params else None
        meta = self._client.head(self._url(endpoint), params=params_).raise_for_status()

        if (meta.headers.get("Transfer-Encoding") == "chunked") or (
            "content-length" not in meta.headers
        ):
            logging.debug("Streaming response")
            return backoff(self._stream, self._url(endpoint), params=params_)

        logging.debug("Non-streaming response")
        return (
            backoff(self._client.get, self._url(endpoint), params=params_)
            .raise_for_status()
            .json()
        )

    def _stream(
        self,
        endpoint: str,
        *,
        params: QueryParamTypes | None = None,
        method: str = "GET",
    ) -> Any:
        with self._client.stream(method, endpoint, params=params) as stream:
            stream.raise_for_status()
            buff = json.loads("".join(data for data in stream.iter_text()))
        return buff

    def _url(self, endpoint: str) -> str:
        return f"{LuftmessnetzClient.BASE_URL}/{endpoint}"



def backoff(func, *args, **kwargs) -> Any:   # type: ignore
    base = 2
    for i in range(1, MAX_RETRIES + 1):
        try:
            return func(*args, **kwargs)
        except httpx.RemoteProtocolError:
            logging.debug(f"RemoteProtoclolError: retry {i}")
            timeout = base**i
            time.sleep(timeout)
